"""Controller holding and managing HUE resources of type `device`."""
from typing import List, Optional, Type

from ..models.zigbee_connectivity import ZigbeeConnectivity
from ..models.device import Device, DevicePut
from ..models.feature import IdentifyFeature
from ..models.light import Light
from ..models.resource import ResourceTypes
from ..models.room import Room
from .base import BaseResourcesController
from .sensors import SENSOR_TYPES


class DevicesController(BaseResourcesController[Type[Device]]):
    """Controller holding and managing HUE resources of type `device`."""

    item_type = ResourceTypes.DEVICE
    item_cls = Device

    def get_lights(self, id: str) -> List[Light]:
        """Return all lights belonging to given device."""
        return [x for x in self._bridge.lights if x.id in self[id].lights]

    def get_sensors(self, id: str) -> List[SENSOR_TYPES]:
        """Return all sensors belonging to given device."""
        return [x for x in self._bridge.sensors if x.id in self[id].sensors]

    def get_room(self, id: str) -> Optional[Room]:
        """Return room the given device belongs to (if any)."""
        return next((x for x in self._bridge.groups.room if id in x.devices), None)

    def get_zigbee_connectivity(self, id: str) -> Optional[ZigbeeConnectivity]:
        """Return the ZigbeeConnectivity resource connected to device."""
        for service in self._items[id].services:
            if service.rtype == ResourceTypes.ZIGBEE_CONNECTIVITY:
                return self._bridge.sensors.zigbee_connectivity.get(service.rid)
        return None

    async def set_identify(self, id: str) -> None:
        """
        Send identify command to the device.

        Triggers a visual identification sequence, current implemented as
        (which can change in the future):
        - Bridge performs Zigbee LED identification cycles for 5 seconds
        - Lights perform one breathe cycle
        - Sensors perform LED identification cycles for 15 seconds
        """
        await self.update(id, DevicePut(identify=IdentifyFeature()))



"""
Model(s) for matter resource on HUE bridge.

https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_matter
"""
from dataclasses import dataclass
from enum import Enum
from typing import Optional

from .resource import ResourceTypes


class MatterAction(Enum):
    """
    Enum with the possible Matter actions.

    reset: Resets Matter, including removing all fabrics and reset state to factory settings.
    """

    NONE = ""
    RESET = "reset"


@dataclass
class Matter:
    """
    Represent a (full) `Matter` resource when retrieved from the api.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_matter_get
    """

    id: str
    max_fabrics: int  # Maximum number of fabrics that can exist at a time
    has_qr_code: bool  # Indicates whether a physical QR code is present

    id_v1: Optional[str] = None
    type: ResourceTypes = ResourceTypes.MATTER


@dataclass
class MatterPut:
    """
    Matter resource properties that can be set/updated with a PUT request.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_matter__id__put
    """

    action: Optional[MatterAction] = None



"""
Model(s) for behavior_instance resource on HUE bridge.

API to manage instances of script.
https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_behavior_instance
"""
from dataclasses import dataclass
from enum import Enum
from typing import List, Optional

from .resource import ResourceIdentifier, ResourceTypes


class BehaviorInstanceStatus(Enum):
    """Behavior script instance status."""

    INITIALIZING = "initializing"
    RUNNING = "running"
    DISABLED = "disabled"
    ERRORED = "errored"


@dataclass
class BehaviorInstanceMetadata:
    """Represent BehaviorInstance Metadata object as used by BehaviorInstance resource."""

    name: Optional[str] = None


class DependencyLevel(Enum):
    """
    Enum with dependency Levels.

    - non_critical: Defines a dependency between resources:
      although source is impacted by removal of target, source remains
      of target means removal of source.
    - critical: Defines a critical dependency between resources:
      source cannot function without target,
      hence removal of target means removal of source.
    """

    NON_CRITICAL = "non_critical"
    CRITICAL = "critical"


@dataclass
class ResourceDependee:
    """
    ResourceDependee object as used by the Hue api.

    Represents a resource which (this) resource is dependent on.

    clip-api.schema.json#/definitions/ResourceDependeeGet
    """

    # target: required(object - Id of the dependency resource (target).
    target: ResourceIdentifier
    level: DependencyLevel
    type: Optional[str] = None


@dataclass
class BehaviorInstance:
    """
    Represent a (full) `BehaviorInstance` resource when retrieved from the api.

    BehaviorInstance is an instance of a script.
    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_behavior_instance_get
    """

    id: str
    script_id: str
    enabled: bool

    # configuration: required(object)
    # Script configuration.
    # This property is validated using ScriptDefinition.configuration_schema JSON schema.
    configuration: dict

    # dependees: required(array of ResourceDependee)
    # Represents all resources which this instance depends on.
    dependees: List[ResourceDependee]

    # status: required(one of initializing, running, disabled, errored)
    # If the script is in the errored state then check errors for more details about the error.
    status: BehaviorInstanceStatus

    # last_error: required(string) - Last error happened while executing the script.
    last_error: str
    metadata: BehaviorInstanceMetadata

    # state: (object)
    # Script instance state.
    # This read-only property is according to ScriptDefinition.state_schema JSON schema.
    state: Optional[dict] = None

    id_v1: Optional[str] = None
    migrated_from: Optional[str] = None
    type: ResourceTypes = ResourceTypes.BEHAVIOR_INSTANCE


@dataclass
class BehaviorInstancePut:
    """
    Properties to send when updating/setting a `BehaviorInstance` object on the api.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_behavior_instance__put
    """

    enabled: Optional[bool] = None
    # configuration: (object) Script configuration.
    # This property is validated using ScriptDefinition.configuration_schema JSON schema.
    configuration: Optional[dict] = None
    # trigger: (object) Action that needs to be taken by this script instance.
    # This property is validated using ScriptDefinition.trigger_schema JSON schema.
    trigger: Optional[dict] = None
    metadata: Optional[BehaviorInstanceMetadata] = None


@dataclass
class BehaviorInstancePost:
    """
    Properties to send when creating a `BehaviorInstance` object on the api.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#behavior_instance_post
    """

    script_id: str
    enabled: bool
    configuration: dict
    metadata: Optional[BehaviorInstanceMetadata] = None
    migrated_from: Optional[str] = None



"""
Model(s) for motion resource on HUE bridge.

https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_motion
"""
from dataclasses import dataclass
from typing import Optional

from .resource import ResourceIdentifier, ResourceTypes


@dataclass
class MotionSensingFeature:
    """
    Represent MotionSensingFeature as retrieved from api.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_motion_get
    """

    motion: Optional[bool]  # value is None/null when sensor is disabled!
    motion_valid: bool


@dataclass
class Motion:
    """
    Represent a (full) `Motion` resource when retrieved from the api.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_motion_get
    """

    id: str
    owner: ResourceIdentifier
    # enabled: required(boolean)
    # true when sensor is activated, false when deactivated
    enabled: bool
    motion: MotionSensingFeature

    id_v1: Optional[str] = None
    type: ResourceTypes = ResourceTypes.MOTION


@dataclass
class MotionPut:
    """
    Motion resource properties that can be set/updated with a PUT request.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_motion__id__put
    """

    enabled: Optional[bool] = None



"""
Model(s) for grouped_light resource on HUE bridge.

https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_grouped_light
"""
from dataclasses import dataclass
from typing import Optional

from .feature import (
    AlertFeature,
    AlertFeaturePut,
    ColorFeaturePut,
    ColorTemperatureDeltaFeaturePut,
    ColorTemperatureFeaturePut,
    DimmingDeltaFeaturePut,
    DimmingFeatureBase,
    DynamicsFeaturePut,
    OnFeature,
)
from .resource import ResourceTypes


@dataclass
class GroupedLight:
    """
    Represent a (full) GroupedLight object when retrieved from the api.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_grouped_light_get
    """

    id: str
    id_v1: Optional[str] = None
    on: Optional[OnFeature] = None
    dimming: Optional[DimmingFeatureBase] = None
    alert: Optional[AlertFeature] = None
    type: ResourceTypes = ResourceTypes.GROUPED_LIGHT


@dataclass
class GroupedLightPut:
    """
    Represent a GroupedLight model when sending a PUT request.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_grouped_light__id__put
    """

    on: Optional[OnFeature] = None
    dimming: Optional[DimmingFeatureBase] = None
    dimming_delta: Optional[DimmingDeltaFeaturePut] = None
    color_temperature: Optional[ColorTemperatureFeaturePut] = None
    color_temperature_delta: Optional[ColorTemperatureDeltaFeaturePut] = None
    color: Optional[ColorFeaturePut] = None
    dynamics: Optional[DynamicsFeaturePut] = None
    alert: Optional[AlertFeaturePut] = None



"""
Model(s) for geofence_client resource on HUE bridge.

https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_geofence_client
"""
from dataclasses import dataclass
from typing import Optional

from .resource import ResourceTypes


@dataclass
class GeofenceClient:
    """
    Represent a (full) `GeofenceClient` resource when retrieved from the api.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_geofence_client_get
    """

    id: str
    name: str
    id_v1: Optional[str] = None
    type: ResourceTypes = ResourceTypes.GEOFENCE_CLIENT


@dataclass
class GeofenceClientPut:
    """
    GeofenceClient resource properties that can be set/updated with a PUT request.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_geofence_client__id__put
    """

    is_at_home: Optional[bool] = None
    name: Optional[str] = None


@dataclass
class GeofenceClientPost:
    """
    GeofenceClient resource properties that can be set with a POST request.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_geofence_client_post
    """

    is_at_home: Optional[bool] = None
    name: Optional[str] = None
    type: ResourceTypes = ResourceTypes.GEOFENCE_CLIENT



"""Base controller for HUE resources as retrieved from the Hue bridge."""
import asyncio
from asyncio.coroutines import iscoroutinefunction

from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Generic,
    Iterator,
    List,
    Optional,
    Tuple,
    TypeVar,
    Union,
)

from ..models.device import Device

from ...util import NoneType, dataclass_from_dict, dataclass_to_dict, update_dataclass
from ..models.resource import ResourceTypes
from .events import EventCallBackType, EventType

if TYPE_CHECKING:
    from .. import HueBridgeV2


EventSubscriptionType = Tuple[
    EventCallBackType,
    "Optional[Tuple[EventType]]",
]

ID_FILTER_ALL = "*"

CLIPResource = TypeVar("CLIPResource")


class BaseResourcesController(Generic[CLIPResource]):
    """Holds and manages all items for a specific Hue resource type."""

    item_type: Optional[ResourceTypes] = None
    item_cls: Optional[CLIPResource] = None
    allow_parser_error = False

    def __init__(self, bridge: "HueBridgeV2") -> None:
        """Initialize instance."""
        self._bridge = bridge
        self._items: Dict[str, CLIPResource] = {}
        self._logger = bridge.logger.getChild(self.item_type.value)
        self._subscribers: Dict[str, EventSubscriptionType] = {ID_FILTER_ALL: []}
        self._initialized = False

    @property
    def items(self) -> List[CLIPResource]:
        """Return all items for this resource."""
        return list(self._items.values())

    async def initialize(self, initial_data: Optional[List[dict]]) -> None:
        """
        Initialize controller by fetching all items for this resource type from bridge.

        Must be called only once. Any updates will be retrieved by events.
        Optionally provide the initial data.
        """
        if initial_data is None:
            endpoint = f"clip/v2/resource/{self.item_type.value}"
            initial_data = await self._bridge.request("get", endpoint)
        else:
            initial_data = [
                x for x in initial_data if x["type"] == self.item_type.value
            ]
        self._logger.debug("fetched %s items", len(initial_data))

        if self._initialized:
            # we're already initialized, treat this as reconnect
            await self.__handle_reconnect(initial_data)
            return

        for item in initial_data:
            await self._handle_event(EventType.RESOURCE_ADDED, item)
        # subscribe to item updates
        self._bridge.events.subscribe(
            self._handle_event, resource_filter=self.item_type
        )
        self._initialized = True

    def subscribe(
        self,
        callback: EventCallBackType,
        id_filter: Union[str, Tuple[str], None] = None,
        event_filter: Union[EventType, Tuple[EventType], None] = None,
    ) -> Callable:
        """
        Subscribe to status changes for this resource type.

        Parameters:
            - `callback` - callback function to call when an event emits.
            - `id_filter` - Optionally provide resource ID(s) to filter events for.
            - `event_filter` - Optionally provide EventType(s) as filter.

        Returns:
            function to unsubscribe.
        """
        if not isinstance(event_filter, (NoneType, (list, tuple))):
            event_filter = (event_filter,)

        if id_filter is None:
            id_filter = (ID_FILTER_ALL,)
        elif not isinstance(id_filter, (list, tuple)):
            id_filter = (id_filter,)

        subscription = (callback, event_filter)

        for id_key in id_filter:
            if id_key not in self._subscribers:
                self._subscribers[id_key] = []
            self._subscribers[id_key].append(subscription)

        # unsubscribe logic
        def unsubscribe():
            for id_key in id_filter:
                if id_key not in self._subscribers:
                    continue
                self._subscribers[id_key].remove(subscription)

        return unsubscribe

    def get_by_v1_id(self, id: str) -> Optional[CLIPResource]:
        """Get item by it's legacy V1 id."""
        return next((item for item in self._items.values() if item.id_v1 == id), None)

    def get_device(self, id: str) -> Optional[Device]:
        """
        Return device the given resource belongs to.

        Returns None if the resource id is (no longer) valid
        or does not belong to a device.
        """
        if self.item_type == ResourceTypes.DEVICE:
            return self[id]
        for device in self._bridge.devices:
            for service in device.services:
                if service.rid == id:
                    return device
        return None

    async def update(self, id: str, obj_in: CLIPResource) -> None:
        """
        Update HUE resource with PUT command.

        Provide instance of object's class with only the changed keys set.
        Note that not all resources allow updating/setting of data.
        Sending keys that are not allowed, results in an error from the bridge.
        """
        endpoint = f"clip/v2/resource/{self.item_type.value}/{id}"
        # create a clean dict with only the changed keys set.
        data = dataclass_to_dict(obj_in, skip_none=True)
        await self._bridge.request("put", endpoint, json=data)

    async def create(self, obj_in: CLIPResource) -> None:
        """
        Create HUE resource with POST command.

        Provide instance of object's class with only the required/allowed keys set.
        Note that not all resources allow creating of items.
        Sending keys that are not allowed, results in an error from the bridge.
        """
        endpoint = f"clip/v2/resource/{self.item_type.value}"
        # create a clean dict with only the not None keys set.
        data = dataclass_to_dict(obj_in, skip_none=True)
        await self._bridge.request("post", endpoint, json=data)

    def get(self, id: str, default: Any = None) -> Optional[CLIPResource]:
        """Get item by id of default if item does not exist."""
        return self._items.get(id, default)

    def __getitem__(self, id: str) -> CLIPResource:
        """Get item by id."""
        return self._items[id]

    def __iter__(self) -> Iterator[CLIPResource]:
        """Iterate items."""
        return iter(self._items.values())

    def __contains__(self, id: str) -> bool:
        """Return bool if id is in items."""
        return id in self._items

    async def _handle_event(
        self, evt_type: EventType, evt_data: Optional[dict], is_reconnect: bool = False
    ) -> None:
        """Handle incoming event for this resource from the EventStream."""
        if evt_data is None:
            return
        item_id = evt_data.get("rid", evt_data["id"])
        if evt_type == EventType.RESOURCE_ADDED:
            # new item added
            try:
                cur_item = self._items[item_id] = dataclass_from_dict(
                    self.item_cls, evt_data
                )
            except (KeyError, ValueError, TypeError) as exc:
                # In an attempt to not completely crash when a single resource can't be parsed
                # due to API schema mismatches, bugs in Hue or other factors, we allow some
                # resources to be skipped. This only works for resources that are not dependees
                # for other resources so we define this in the controller level.
                if not self.allow_parser_error:
                    raise exc
                self._logger.error(
                    "Unable to parse resource, please report this to the authors of aiohue.",
                    exc_info=exc,
                )
                return
        elif evt_type == EventType.RESOURCE_DELETED:
            # existing item deleted
            cur_item = self._items.pop(item_id, evt_data)
        elif evt_type == EventType.RESOURCE_UPDATED:
            # existing item updated
            cur_item = self._items.get(item_id)
            if cur_item is None:
                # should not be possible but just in case
                # if this does happen often we should consider fetching the full object
                self._logger.warning("received update for unknown item %s", item_id)
                return
            # update the existing data with the changed keys/data
            updated_keys = update_dataclass(cur_item, evt_data)
            # do not forward update event at reconnects if no keys were updated
            if len(updated_keys) == 0 and is_reconnect:
                return
            # Do not forward update events for button resource if
            # the button feature is missing in event data in an attempt to prevent
            # ghost events at bridge reboots/firmware updates.
            # in fact this is a feature request to Signify to handle these stateless
            # device events in a different way:
            # https://developers.meethue.com/forum/t/differentiate-stateless-events/6627
            if self.item_type == ResourceTypes.BUTTON and not evt_data.get("button"):
                return
            if self.item_type == ResourceTypes.RELATIVE_ROTARY and not evt_data.get(
                "relative_rotary"
            ):
                return
        else:
            # ignore all other events
            return

        subscribers = (
            self._subscribers.get(item_id, []) + self._subscribers[ID_FILTER_ALL]
        )
        for (callback, event_filter) in subscribers:
            if event_filter is not None and evt_type not in event_filter:
                continue
            # dispatch the full resource object to the callback
            if iscoroutinefunction(callback):
                asyncio.create_task(callback(evt_type, cur_item))
            else:
                callback(evt_type, cur_item)

    async def __handle_reconnect(self, full_state: List[dict]) -> None:
        """Force update of state (on reconnect)."""
        # When the connection to the eventstream was lost for a longer time
        # we fetch the full state to check what changed
        # NOTE: Short connection drops are handled by the `Last-Event-Id` mechanism.
        prev_ids = set(self._items.keys())
        cur_ids = set()
        for item in full_state:
            cur_ids.add(item["id"])
            if item["id"] not in prev_ids:
                # item added
                await self._handle_event(EventType.RESOURCE_ADDED, item)
            else:
                # work out if the item changed in the regular event logic
                # ignore stateless (button) resources to prevent false positive state events
                if item["type"] in (
                    ResourceTypes.BUTTON.value,
                    ResourceTypes.RELATIVE_ROTARY.value,
                ):
                    continue
                await self._handle_event(
                    EventType.RESOURCE_UPDATED, item, is_reconnect=True
                )

        # work out item deletions
        deleted_ids = {x for x in prev_ids if x not in cur_ids}
        for resource_id in deleted_ids:
            self._handle_event(
                EventType.RESOURCE_DELETED,
                {"rtype": self.item_type, "rid": resource_id},
            )


class GroupedControllerBase(Generic[CLIPResource]):
    """Convenience controller which combines items from multiple resources."""

    def __init__(
        self, bridge: "HueBridgeV2", resources: List[BaseResourcesController]
    ) -> None:
        """Initialize instance."""
        self._resources = resources
        self._bridge = bridge
        self._logger = bridge.logger.getChild(self.__class__.__name__.lower())
        self._subscribers: List[Tuple[EventCallBackType, Optional[str]]] = []

    @property
    def resources(self) -> List[BaseResourcesController]:
        """Return all resource controllers that are grouped by this groupcontroller."""
        return self._resources

    @property
    def items(self) -> List[CLIPResource]:
        """Return all items from all grouped resources."""
        return [x for y in self._resources for x in y]

    def subscribe(
        self,
        callback: EventCallBackType,
        id_filter: Union[str, Tuple[str], None] = None,
        event_filter: Union[EventType, Tuple[EventType], None] = None,
    ) -> Callable:
        """Subscribe to status changes for all grouped resources."""
        unsubs = [
            x.subscribe(callback, id_filter, event_filter) for x in self._resources
        ]

        def unsubscribe():
            for unsub in unsubs:
                unsub()

        return unsubscribe

    async def initialize(self, initial_data: Optional[List[dict]]) -> None:
        """Initialize controller for all watched resources."""
        for resource_control in self._resources:
            await resource_control.initialize(initial_data)

    def get_device(self, id: str) -> Optional[Device]:
        """
        Return device the given resource belongs to.

        Returns None if the resource id is (no longer) valid
        or does not belong to a device.
        """
        for ctrl in self._resources:
            if id in ctrl:
                return ctrl.get_device(id)
        return None

    def get(self, id: str, default: Any = None) -> Optional[CLIPResource]:
        """Get item by id of default if item does not exist."""
        return next((x for y in self._resources for x in y if x.id == id), default)

    def __getitem__(self, id: str) -> CLIPResource:
        """Get item by id."""
        return next((x for y in self._resources for x in y if x.id == id))

    def __iter__(self) -> Iterator[CLIPResource]:
        """Iterate items."""
        return iter(self.items)

    def __contains__(self, id: str) -> bool:
        """Return bool if id is in items."""
        return any((x for x in self.items if x.id == id))



"""
Model(s) for zigbee_connectivity resource on HUE bridge.

https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_zigbee_connectivity
"""
from dataclasses import dataclass
from enum import Enum


from typing import Optional

from .resource import ResourceIdentifier, ResourceTypes


class ConnectivityServiceStatus(Enum):
    """
    Enum with possible ConnectivityService statuses.

    Connected if device has been recently been available.
    When indicating connectivity issues the device is powered off or has network
    issues When indicating unidirectional incoming the device only talks to bridge
    """

    CONNECTED = "connected"
    DISCONNECTED = "disconnected"
    CONNECTIVITY_ISSUE = "connectivity_issue"
    UNIDIRECTIONAL_INCOMING = "unidirectional_incoming"


@dataclass
class ZigbeeConnectivity:
    """
    Represent a (full) `ZigbeeConnectivity` resource when retrieved from the api.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_zigbee_connectivity_get
    """

    id: str
    owner: ResourceIdentifier
    status: ConnectivityServiceStatus
    mac_address: str
    id_v1: Optional[str] = None
    type: ResourceTypes = ResourceTypes.ZIGBEE_CONNECTIVITY



"""
Model(s) for temperature resource on HUE bridge.

https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_temperature
"""

from dataclasses import dataclass
from typing import Optional

from .resource import ResourceIdentifier, ResourceTypes


@dataclass
class TemperatureFeature:
    """Represent TemperatureFeature."""

    temperature: float
    temperature_valid: bool


@dataclass
class Temperature:
    """
    Represent a (full) `Temperature` resource when retrieved from the api.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_temperature_get
    """

    id: str
    owner: ResourceIdentifier
    enabled: bool
    temperature: TemperatureFeature

    id_v1: Optional[str] = None
    type: ResourceTypes = ResourceTypes.TEMPERATURE


@dataclass
class TemperaturePut:
    """
    Temperature resource properties that can be set/updated with a PUT request.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_temperature__id__put
    """

    enabled: Optional[bool] = None



"""Aiohue errors.

https://developers.meethue.com/documentation/error-messages
"""


class AiohueException(Exception):
    """Base exception for aiohue."""


class Unauthorized(AiohueException):
    """Username is not authorized."""


class LinkButtonNotPressed(AiohueException):
    """Raised when trying to create a user but link button not pressed."""


class InvalidEvent(AiohueException):
    """Raised when we receive an event that we can not (yet) handle."""


class InvalidAPIVersion(AiohueException):
    """Raised when we're trying to connect to an unsupported bridge version."""


class BridgeBusy(AiohueException):
    """Raised when multiple requests to the bridge failed."""


class BridgeSoftwareOutdated(AiohueException):
    """Raised when the software version of the bridge is (too) outdated."""


ERRORS = {1: Unauthorized, 101: LinkButtonNotPressed}


def raise_from_error(error: dict):
    """Raise Exception based on Hue error."""
    type = error.get("type")
    cls = ERRORS.get(type, AiohueException)
    raise cls(error["description"])



from __future__ import annotations
from .api import APIItems


class Scenes(APIItems):
    """Represents Hue Scenes.

    https://developers.meethue.com/documentation/scenes-api
    """

    def __init__(self, logger, raw, request):
        super().__init__(logger, raw, request, "scenes", Scene)


class Scene:
    """Represents a Hue Scene."""

    ITEM_TYPE = "scenes"

    def __init__(self, id, raw, request):
        self.id = id
        self.raw = raw
        self._request = request

    @property
    def name(self):
        return self.raw["name"]

    @property
    def lights(self):
        return self.raw["lights"]

    @property
    def owner(self):
        return self.raw["owner"]

    @property
    def recycle(self):
        return self.raw["recycle"]

    @property
    def locked(self):
        return self.raw["locked"]

    @property
    def appdata(self):
        return self.raw["appdata"]

    @property
    def picture(self):
        return self.raw["picture"]

    @property
    def lastupdated(self):
        return self.raw["lastupdated"]

    @property
    def version(self):
        return self.raw["version"]



"""Controller holding and managing HUE resources of sensor type."""
import asyncio
from typing import TYPE_CHECKING, Dict, Optional, Type, Union

from aiohue.util import dataclass_to_dict

from ..models.button import Button, ButtonEvent
from ..models.relative_rotary import RelativeRotary
from ..models.device_power import DevicePower
from ..models.geofence_client import GeofenceClient
from ..models.light_level import LightLevel, LightLevelPut
from ..models.motion import Motion, MotionPut
from ..models.resource import ResourceTypes
from ..models.temperature import Temperature
from ..models.zigbee_connectivity import ZigbeeConnectivity
from .base import BaseResourcesController, GroupedControllerBase
from .events import EventType

if TYPE_CHECKING:
    from .. import HueBridgeV2

SENSOR_TYPES = Union[
    DevicePower,
    Button,
    GeofenceClient,
    LightLevel,
    Motion,
    RelativeRotary,
    Temperature,
    ZigbeeConnectivity,
]

BTN_WORKAROUND_NEEDED = ("FOHSWITCH",)


class DevicePowerController(BaseResourcesController[Type[DevicePower]]):
    """Controller holding and managing HUE resources of type `device_power`."""

    item_type = ResourceTypes.DEVICE_POWER
    item_cls = DevicePower
    allow_parser_error = True


class ButtonController(BaseResourcesController[Type[Button]]):
    """Controller holding and managing HUE resources of type `button`."""

    item_type = ResourceTypes.BUTTON
    item_cls = Button
    allow_parser_error = True

    _workaround_tasks: Dict[str, asyncio.Task] = None

    async def _handle_event(
        self, evt_type: EventType, evt_data: Optional[dict]
    ) -> None:
        """Handle incoming event for this resource from the EventStream."""
        await super()._handle_event(evt_type, evt_data)

        if not evt_data:
            return

        # Handle longpress workaround if needed
        if not (
            evt_type == EventType.RESOURCE_UPDATED
            and evt_data.get("button", {}).get("last_event")
            == ButtonEvent.INITIAL_PRESS.value
        ):
            return

        device = self.get_device(evt_data["id"])
        if device is None or device.product_data.model_id not in BTN_WORKAROUND_NEEDED:
            return

        if self._workaround_tasks is None:
            self._workaround_tasks = {}

        # cancel existing task (if any)
        # should not happen, but just in case
        task = self._workaround_tasks.pop(evt_data["id"], None)
        if task and not task.done():
            task.cancel()

        self._workaround_tasks[evt_data["id"]] = asyncio.create_task(
            self._handle_longpress_workaround(evt_data["id"])
        )

    async def _handle_longpress_workaround(self, id: int):
        """Handle workaround for FOH switches."""
        # Fake `held down` and `long press release` events.
        # This might need to be removed in a future release once/if Signify
        # adds this back in their API.
        self._logger.debug("Long press workaround for FOH switch initiated.")
        btn_resource = dataclass_to_dict(self._items[id])
        await asyncio.sleep(1.5)  # time to initially wait for SHORT_RELEASE
        count = 0
        try:
            while count <= 20:  # = max 10 seconds
                cur_event = self._items[id].button.last_event
                if cur_event == ButtonEvent.SHORT_RELEASE:
                    break
                # send REPEAT until short release is received
                btn_resource["button"]["last_event"] = ButtonEvent.REPEAT.value
                await self._handle_event(EventType.RESOURCE_UPDATED, btn_resource)
                await asyncio.sleep(0.5)
                count += 1
        finally:
            # Fire LONG_RELEASE event if time between INITIAL_PRESS and SHORT_RELEASE
            # is more than 1.5 seconds or 10 seconds expired.
            # The button will not send SHORT_RELEASE if more than 10 seconds passed.
            # Note that the button will also fire the SHORT_RELEASE event if it's released within
            # those 10 seconds.
            if count > 1:
                btn_resource["button"]["last_event"] = ButtonEvent.LONG_RELEASE.value
                await self._handle_event(EventType.RESOURCE_UPDATED, btn_resource)
            self._logger.debug("Long press workaround for FOH switch completed.")


class GeofenceClientController(BaseResourcesController[Type[GeofenceClient]]):
    """Controller holding and managing HUE resources of type `geofence_client`."""

    item_type = ResourceTypes.GEOFENCE_CLIENT
    item_cls = GeofenceClient
    allow_parser_error = True


class LightLevelController(BaseResourcesController[Type[LightLevel]]):
    """Controller holding and managing HUE resources of type `light_level`."""

    item_type = ResourceTypes.LIGHT_LEVEL
    item_cls = LightLevel
    allow_parser_error = True

    async def set_enabled(self, id: str, enabled: bool) -> None:
        """Enable/Disable sensor."""
        await self.update(id, LightLevelPut(enabled=enabled))


class MotionController(BaseResourcesController[Type[Motion]]):
    """Controller holding and managing HUE resources of type `motion`."""

    item_type = ResourceTypes.MOTION
    item_cls = Motion
    allow_parser_error = True

    async def set_enabled(self, id: str, enabled: bool) -> None:
        """Enable/Disable sensor."""
        await self.update(id, MotionPut(enabled=enabled))


class RelativeRotaryController(BaseResourcesController[Type[Button]]):
    """Controller holding and managing HUE resources of type `relative_rotary`."""

    item_type = ResourceTypes.RELATIVE_ROTARY
    item_cls = RelativeRotary
    allow_parser_error = True


class TemperatureController(BaseResourcesController[Type[Temperature]]):
    """Controller holding and managing HUE resources of type `temperature`."""

    item_type = ResourceTypes.TEMPERATURE
    item_cls = Temperature
    allow_parser_error = True


class ZigbeeConnectivityController(BaseResourcesController[Type[ZigbeeConnectivity]]):
    """Controller holding and managing HUE resources of type `zigbee_connectivity`."""

    item_type = ResourceTypes.ZIGBEE_CONNECTIVITY
    item_cls = ZigbeeConnectivity
    allow_parser_error = True


class SensorsController(GroupedControllerBase[SENSOR_TYPES]):
    """Controller grouping resources of all sensor resources."""

    def __init__(self, bridge: "HueBridgeV2") -> None:
        """Initialize instance."""
        self.button = ButtonController(bridge)
        self.device_power = DevicePowerController(bridge)
        self.geofence_client = GeofenceClientController(bridge)
        self.light_level = LightLevelController(bridge)
        self.motion = MotionController(bridge)
        self.relative_rotary = RelativeRotaryController(bridge)
        self.temperature = TemperatureController(bridge)
        self.zigbee_connectivity = ZigbeeConnectivityController(bridge)
        super().__init__(
            bridge,
            [
                self.button,
                self.device_power,
                self.geofence_client,
                self.light_level,
                self.motion,
                self.relative_rotary,
                self.temperature,
                self.zigbee_connectivity,
            ],
        )



"""
Model(s) for button resource on HUE bridge.

https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_button
"""
from dataclasses import dataclass
from enum import Enum
from typing import Optional, Type

from .resource import ResourceIdentifier, ResourceTypes


class ButtonEvent(Enum):
    """Enum with possible button events."""

    INITIAL_PRESS = "initial_press"
    REPEAT = "repeat"
    SHORT_RELEASE = "short_release"
    LONG_RELEASE = "long_release"
    DOUBLE_SHORT_RELEASE = "double_short_release"
    UNKNOWN = "unknown"

    @classmethod
    def _missing_(cls: Type, value: object):
        """Set default enum member if an unknown value is provided."""
        return ButtonEvent.UNKNOWN


@dataclass
class ButtonFeature:
    """Represent ButtonFeature object as used by the Hue api."""

    last_event: ButtonEvent


@dataclass
class ButtonMetadata:
    """Represent ButtonMetadata object as used by the Button resource."""

    # number of control within switch (value between 0..8).
    # Meaning in combination with type
    # - dots: Number of dots
    # - number: Number printed on device
    # - other: a logical order of controls in switch
    control_id: int


@dataclass
class Button:
    """
    Represent a (full) `Button` resource when retrieved from the api.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_button_get
    """

    id: str
    owner: ResourceIdentifier
    metadata: ButtonMetadata

    button: Optional[ButtonFeature] = None
    id_v1: Optional[str] = None
    type: ResourceTypes = ResourceTypes.BUTTON



"""Various helper utility for Hue bridge discovery."""
from __future__ import annotations

from dataclasses import dataclass
from typing import List

from aiohttp import ClientConnectionError, ClientError, ClientSession

from .util import normalize_bridge_id

URL_NUPNP = "https://discovery.meethue.com/"


@dataclass(frozen=True)
class DiscoveredHueBridge:
    """Model for a discovered Hue bridge."""

    host: str
    id: str
    supports_v2: bool


async def discover_bridge(
    host: str,
    websession: ClientSession | None = None,
) -> DiscoveredHueBridge:
    """
    Discover bridge details from given hostname/ip.

    Raises exception from aiohttp if there is no bridge alive on given ip/host.
    """
    websession_provided = websession is not None
    if websession is None:
        websession = ClientSession()
    try:
        bridge_id = await is_hue_bridge(host, websession)
        supports_v2 = await is_v2_bridge(host, websession)
        return DiscoveredHueBridge(host, bridge_id, supports_v2)
    finally:
        if not websession_provided:
            await websession.close()


async def discover_nupnp(
    websession: ClientSession | None = None,
) -> List[DiscoveredHueBridge]:
    """Discover bridges via NUPNP."""
    result = []
    websession_provided = websession is not None
    if websession is None:
        websession = ClientSession()
    try:
        async with websession.get(URL_NUPNP, timeout=30) as res:
            for item in await res.json():
                host = item["internalipaddress"]
                # the nupnp discovery might return items that are not in local network
                # connect to each bridge to find out if it's alive.
                try:
                    result.append(await discover_bridge(host, websession))
                except Exception:  # noqa
                    pass
        return result
    except ClientError:
        return result
    finally:
        if not websession_provided:
            await websession.close()


async def is_hue_bridge(host: str, websession: ClientSession | None = None) -> str:
    """
    Check if there is a bridge alive on given ip and return bridge ID.

    Raises exception from aiohttp if the bridge can not be reached on given hostname.
    """
    websession_provided = websession is not None
    if websession is None:
        websession = ClientSession()
    try:
        # every hue bridge returns discovery info on this endpoint
        url = f"http://{host}/api/config"
        async with websession.get(url, timeout=30) as res:
            res.raise_for_status()
            data = await res.json()
            if "bridgeid" not in data:
                # there are some emulator projects out there that emulate a Hue bridge
                # in a sloppy way, ignore them.
                # https://github.com/home-assistant-libs/aiohue/issues/134
                raise ClientConnectionError(
                    "Invalid API response, not a real Hue bridge?"
                )
            return normalize_bridge_id(data["bridgeid"])
    finally:
        if not websession_provided:
            await websession.close()


async def is_v2_bridge(host: str, websession: ClientSession | None = None) -> bool:
    """Check if the bridge has support for the new V2 api."""
    websession_provided = websession is not None
    if websession is None:
        websession = ClientSession()
    try:
        # v2 api is https only and returns a 403 forbidden when no key provided
        url = f"https://{host}/clip/v2/resource"
        async with websession.get(
            url, ssl=False, raise_for_status=False, timeout=30
        ) as res:
            return res.status == 403
    except Exception:  # pylint: disable=broad-except
        # all other status/exceptions means the bridge is not v2 or not reachable at this time
        return False
    finally:
        if not websession_provided:
            await websession.close()



"""
Model(s) for Scene resource on HUE bridge.

https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_scene
"""
from dataclasses import dataclass
from typing import List, Optional

from .feature import (
    ColorFeatureBase,
    ColorTemperatureFeatureBase,
    DimmingFeatureBase,
    DynamicsFeaturePut,
    GradientFeatureBase,
    OnFeature,
    PaletteFeature,
    RecallFeature,
    SceneEffectsFeature,
)
from .resource import ResourceIdentifier, ResourceTypes


@dataclass
class ActionAction:
    """Represent (scene) `ActionAction` model."""

    on: Optional[OnFeature] = None
    dimming: Optional[DimmingFeatureBase] = None
    color: Optional[ColorFeatureBase] = None
    color_temperature: Optional[ColorTemperatureFeatureBase] = None
    gradient: Optional[GradientFeatureBase] = None
    effects: Optional[SceneEffectsFeature] = None
    dynamics: Optional[DynamicsFeaturePut] = None


@dataclass
class Action:
    """Represent (scene) `Action` model."""

    target: ResourceIdentifier
    action: ActionAction


@dataclass
class SceneMetadata:
    """Represent SceneMetadata object as used by the Hue api."""

    name: str
    image: Optional[ResourceIdentifier] = None


@dataclass
class SceneMetadataPut:
    """Represent SceneMetadata model when sent/updated to the API with PUT request."""

    name: str


@dataclass
class Scene:
    """
    Represent (full) `Scene` Model when retrieved from the API.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_scene__id__get
    """

    id: str
    metadata: SceneMetadata
    # group: required(object)
    # Group associated with this Scene. All services in the group are part of this scene.
    # If the group is changed the scene is updated (e.g. light added/removed)
    group: ResourceIdentifier
    # actions: required(array of Action)
    # List of actions to be executed synchronously on recal
    actions: List[Action]
    palette: PaletteFeature
    # speed: required(number – minimum: 0 – maximum: 1)
    speed: float
    # auto_dynamic: whether to automatically start the scene dynamically on active recall
    auto_dynamic: bool

    # optional params
    id_v1: Optional[str] = None

    type: ResourceTypes = ResourceTypes.SCENE


@dataclass
class ScenePut:
    """
    Properties to send when updating/setting a `Scene` object on the api.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_scene__id__put
    """

    metadata: Optional[SceneMetadataPut] = None
    actions: Optional[List[Action]] = None
    palette: Optional[PaletteFeature] = None
    recall: Optional[RecallFeature] = None
    palette: Optional[PaletteFeature] = None
    # speed: (number – minimum: 0 – maximum: 1)
    # Speed of dynamic palette for this scene
    speed: Optional[float] = None
    auto_dynamic: Optional[bool] = None


@dataclass
class ScenePost:
    """
    Properties to send when creating a `Scene` object on the api.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_scene_post
    """

    metadata: SceneMetadata
    group: ResourceIdentifier
    actions: List[Action]
    palette: Optional[PaletteFeature] = None
    speed: Optional[float] = None
    auto_dynamic: Optional[bool] = None
    type: ResourceTypes = ResourceTypes.SCENE



"""
Model(s) for bridge resource on HUE bridge.

https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_bridge
"""
from dataclasses import dataclass
from typing import Optional

from .resource import ResourceTypes


@dataclass
class TimeZone:
    """Represent TimeZone object as received from API."""

    time_zone: str  # e.g. Europe/Amsterdam


@dataclass
class Bridge:
    """
    Represent a (full) `Bridge` resource when retrieved from the api.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_bridge_get
    """

    id: str
    bridge_id: str
    time_zone: TimeZone

    id_v1: Optional[str] = None
    type: ResourceTypes = ResourceTypes.BRIDGE



from __future__ import annotations


class APIItems:
    """Base class for a map of API Items."""

    def __init__(self, logger, raw, request, path, item_cls):
        self._logger = logger
        self._request = request
        self._path = path
        self._item_cls = item_cls
        self._items = {}
        self._process_raw(raw)

    @property
    def items(self):
        return self.values()

    async def update(self):
        """Request update of all items of this resource type."""
        raw = await self._request("get", self._path)
        self._process_raw(raw)

    def _process_raw(self, raw):
        for id, raw_item in raw.items():
            obj = self._items.get(id)

            if obj is not None:
                obj.raw = raw_item
            else:
                self._items[id] = self._item_cls(
                    id,
                    raw_item,
                    self._request,
                )

        removed_items = []

        for id in self._items:
            if id not in raw:
                removed_items.append(id)

        for id in removed_items:
            del self._items[id]

    def values(self):
        return list(self._items.values())

    def __getitem__(self, obj_id):
        return self._items[obj_id]

    def __iter__(self):
        return iter(self._items)



"""
Model(s) for Room resource on HUE bridge.

https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_room
"""
from dataclasses import dataclass
from enum import Enum
from typing import List, Optional, Set, Type


from .resource import ResourceIdentifier, ResourceTypes


class RoomArchetype(Enum):
    """Possible archetypes of a room."""

    LIVING_ROOM = "living_room"
    KITCHEN = "kitchen"
    DINING = "dining"
    BEDROOM = "bedroom"
    KIDS_BEDROOM = "kids_bedroom"
    BATHROOM = "bathroom"
    NURSERY = "nursery"
    RECREATION = "recreation"
    OFFICE = "office"
    GYM = "gym"
    HALLWAY = "hallway"
    TOILET = "toilet"
    FRONT_DOOR = "front_door"
    GARAGE = "garage"
    TERRACE = "terrace"
    GARDEN = "garden"
    DRIVEWAY = "driveway"
    CARPORT = "carport"
    HOME = "home"
    DOWNSTAIRS = "downstairs"
    UPSTAIRS = "upstairs"
    TOP_FLOOR = "top_floor"
    ATTIC = "attic"
    GUEST_ROOM = "guest_room"
    STAIRCASE = "staircase"
    LOUNGE = "lounge"
    MAN_CAVE = "man_cave"
    COMPUTER = "computer"
    STUDIO = "studio"
    MUSIC = "music"
    TV = "tv"
    READING = "reading"
    CLOSET = "closet"
    STORAGE = "storage"
    LAUNDRY_ROOM = "laundry_room"
    BALCONY = "balcony"
    PORCH = "porch"
    BARBECUE = "barbecue"
    POOL = "pool"
    OTHER = "other"

    @classmethod
    def _missing_(cls: Type, value: object):
        """Set default enum member if an unknown value is provided."""
        return RoomArchetype.OTHER


@dataclass
class RoomMetaData:
    """Represent MetaData for a room resource."""

    archetype: RoomArchetype
    name: str


@dataclass
class RoomMetaDataPut:
    """
    Represent Room MetaData properties on update/PUT.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_device__id__put
    """

    archetype: Optional[RoomArchetype]
    name: Optional[str]


@dataclass
class Room:
    """
    Represent a (full) `Room` object as retrieved from the Hue api.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_room_get
    """

    id: str
    # services: required(array of ResourceIdentifier)
    # References all services aggregating control and state of children in the group
    # This includes all services grouped in the group hierarchy given by child relation
    # This includes all services of a device grouped in the group hierarchy given by child relation
    # Aggregation is per service type, ie every service type which can be grouped has a
    # corresponding definition of grouped type
    # Supported types “light”
    services: List[ResourceIdentifier]
    metadata: RoomMetaData

    # children: required(array of ResourceIdentifier)
    # Devices to group by the Room Following children are allowed: device
    children: List[ResourceIdentifier]

    id_v1: Optional[str] = None
    type: ResourceTypes = ResourceTypes.ROOM

    @property
    def devices(self) -> Set[str]:
        """Return set of device id's that belong to this room."""
        return {x.rid for x in self.children}

    @property
    def grouped_light(self) -> Optional[str]:
        """Return the grouped light id that is connected to this room (if any)."""
        if not self.services:
            return None
        return next(
            (x.rid for x in self.services if x.rtype == ResourceTypes.GROUPED_LIGHT),
            None,
        )


@dataclass
class RoomPut:
    """
    Properties to send when updating/setting a `Room` object on the api.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_room__id__put
    """

    children: Optional[List[ResourceIdentifier]] = None
    metadata: Optional[RoomMetaDataPut] = None


@dataclass
class RoomPost:
    """
    Properties to send when creating a `Room` object on the api.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_room_post
    """

    children: List[ResourceIdentifier]
    metadata: RoomMetaData



"""Feature Schemas used by various Hue resources."""
from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum
from typing import List, Optional, Type


@dataclass
class OnFeature:
    """Represent `On` Feature object as used by various Hue resources."""

    on: bool


@dataclass
class DimmingFeatureBase:
    """
    Represent `Dimming` Feature base properties.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_light__id__put
    """

    # brightness: (number - maximum: 100)
    # Brightness percentage. value cannot be 0, writing 0 changes it to lowest possible brightness
    brightness: float


@dataclass
class DimmingFeature(DimmingFeatureBase):
    """Represent `Dimming` Feature object as used by various Hue resources."""

    # Percentage of the maximum lumen the device outputs on minimum brightness
    min_dim_level: Optional[float] = None


@dataclass
class DimmingFeaturePut(DimmingFeatureBase):
    """Represent `Dimming` Feature when updating/sending in PUT requests."""


class DeltaAction(Enum):
    """Enum with delta actions for DimmingDelta and ColorDelta feature."""

    UP = "up"
    DOWN = "down"
    STOP = "stop"


@dataclass
class DimmingDeltaFeaturePut:
    """
    Represent `DimmingDelta` Feature when updating/sending in PUT requests.

    Brightness percentage of full-scale increase delta to current dimlevel. Clip at Max-level or Min-level.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_light__id__put
    """

    action: DeltaAction
    brightness_delta: Optional[float] = None


@dataclass
class ColorTemperatureDeltaFeaturePut:
    """
    Represent `DimmingDelta` Feature when updating/sending in PUT requests.

    Mirek delta to current mirek. Clip at mirek_minimum and mirek_maximum of mirek_schema.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_light__id__put
    """

    action: DeltaAction
    mirek_delta: Optional[int] = None


@dataclass
class Position:
    """
    Represent Position object as used by the Hue api.

    Each Coordinate value is a float between -1 and 1.
    """

    x: float
    y: float
    z: float


class AlertEffectType(Enum):
    """Enum with possible alert effect values."""

    BREATHE = "breathe"
    UNKNOWN = "unknown"

    @classmethod
    def _missing_(cls: Type, value: object):
        """Set default enum member if an unknown value is provided."""
        return AlertEffectType.UNKNOWN


@dataclass
class AlertFeature:
    """Represent AlertFeature object when retrieved from the Hue API."""

    action_values: List[AlertEffectType]


@dataclass
class AlertFeaturePut:
    """Represent AlertFeature object when sent to the Hue API."""

    action: AlertEffectType


@dataclass
class IdentifyFeature:
    """
    Represent IdentifyFeature object as used by the Hue api.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_device__id__put
    identify: Triggers a visual identification sequence, current implemented as
    (which can change in the future):
    Bridge performs Zigbee LED identification cycles for 5 seconds Lights perform one breathe
    cycle Sensors perform LED identification cycles for 15 seconds
    """

    # only used in PUT requests on device
    action: str = "identify"


@dataclass
class ColorPoint:
    """
    CIE XY gamut position.

    Each position value is a float between -1 and 1.
    clip-api.schema.json#/definitions/ColorPoint
    """

    x: float
    y: float


@dataclass
class ColorGamut:
    """
    Color gamut of color bulb.

    Some bulbs do not properly return the Gamut information. In this case this is not present.
    clip-api.schema.json#/definitions/ColorGamut
    """

    red: ColorPoint
    green: ColorPoint
    blue: ColorPoint


class GamutType(Enum):
    """
    Enum with possible Gamut types.

    The gammut types supported by hue.
        A: Gamut of early Philips color-only products.
        B: Limited gamut of first Hue color products.
        C: Richer color gamut of Hue white and color ambiance products.
        Other: Color gamut of non-hue products with non-hue gamuts resp w/o gamut.
    """

    A = "A"
    B = "B"
    C = "C"
    OTHER = "other"


@dataclass
class ColorFeatureBase:
    """Represent `Color` Feature base/required properties."""

    xy: ColorPoint


@dataclass
class ColorFeature(ColorFeatureBase):
    """Represent `Color` Feature object as used by various Hue resources."""

    gamut_type: GamutType = GamutType.OTHER
    gamut: Optional[ColorGamut] = None


@dataclass
class ColorFeaturePut(ColorFeatureBase):
    """Represent `Color` Feature when updating/sending in PUT requests."""


@dataclass
class MirekSchema:
    """Represent Mirek schema."""

    mirek_minimum: int = 153
    mirek_maximum: int = 500

    def __post_init__(self):
        """Auto correct invalid values."""
        # Fix for devices that provide wrong info
        if self.mirek_minimum == 0:
            self.mirek_minimum = 153
        if self.mirek_maximum == 65535:
            self.mirek_maximum = 500


@dataclass
class ColorTemperatureFeatureBase:
    """Represent `ColorTemperature` Feature base/required properties."""

    # Color temperature in mirek (153-500) or None if light not in CT spectrum
    mirek: Optional[int]


@dataclass
class ColorTemperatureFeature(ColorTemperatureFeatureBase):
    """Represent `ColorTemperature` Feature object as used by various Hue resources."""

    mirek_schema: MirekSchema
    # mirek_valid will be false if light is currently not in the ct spectrum
    mirek_valid: bool


@dataclass
class ColorTemperatureFeaturePut:
    """Represent `ColorTemperature` Feature when updating/sending in PUT requests."""

    # Color temperature in mirek (153-500)
    mirek: int


class DynamicStatus(Enum):
    """Enum with all possible dynamic statuses."""

    NONE = "none"
    DYNAMIC_PALETTE = "dynamic_palette"
    UNKNOWN = "unknown"

    @classmethod
    def _missing_(cls: Type, value: object):
        """Set default enum member if an unknown value is provided."""
        return DynamicStatus.UNKNOWN


@dataclass
class DynamicsFeature:
    """
    Represent `DynamicsFeature` object as used by various Hue resources.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_light_get
    """

    # speed: required(number – minimum: 0 – maximum: 1)
    # speed of dynamic palette or effect. The speed is valid for the dynamic palette
    # if the status is dynamic_palette  or for the corresponding effect listed in status.
    # In case of status none, the speed is not valid
    speed: float
    # speed_valid: required(boolean)
    # Indicates whether the value presented in speed is valid
    speed_valid: bool
    # status: required(one of dynamic_palette, none)
    # Current status of the lamp with dynamics.
    status: DynamicStatus
    # status_values: required(array of SupportedDynamicStatus)
    # Statuses in which a lamp could be when playing dynamics.
    status_values: List[DynamicStatus] = field(default_factory=list)


@dataclass
class DynamicsFeaturePut:
    """
    Represent `DynamicsFeature` object when sent to the API in PUT requests.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_light__id__put
    """

    # speed: (number – minimum: 0 – maximum: 1)
    # speed of dynamic palette. The speed is valid for the dynamic palette if the status
    # is dynamic_palette or for the corresponding effect listed in status.
    # In case of status none, the speed is not valid
    speed: Optional[float] = None
    # duration: (integer – maximum: 6000000)
    # Duration of a light transition in ms. Accuracy is in 100ms steps.
    duration: Optional[int] = None


class EffectStatus(Enum):
    """Enum with possible effects."""

    NO_EFFECT = "no_effect"
    CANDLE = "candle"
    FIRE = "fire"
    SPARKLE = "sparkle"
    UNKNOWN = "unknown"

    @classmethod
    def _missing_(cls: Type, value: object):
        """Set default enum member if an unknown value is provided."""
        return EffectStatus.UNKNOWN


@dataclass
class SceneEffectsFeature:
    """
    Represent `EffectsFeature` base object as used by scenes.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_scene_get
    """

    effect: EffectStatus


@dataclass
class EffectsFeature:
    """
    Represent `EffectsFeature` object as used by various Hue resources.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_light_get
    """

    status: EffectStatus
    effect_values: List[EffectStatus] = field(default_factory=list)
    status_values: List[EffectStatus] = field(default_factory=list)


@dataclass
class EffectsFeaturePut:
    """
    Represent `EffectsFeature` object when sent to the API in PUT requests.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_light__id__put
    """

    effect: EffectStatus


class TimedEffectStatus(Enum):
    """Enum with possible timed effects."""

    NO_EFFECT = "no_effect"
    SUNRISE = "sunrise"
    UNKNOWN = "unknown"

    @classmethod
    def _missing_(cls: Type, value: object):
        """Set default enum member if an unknown value is provided."""
        return TimedEffectStatus.UNKNOWN


@dataclass
class TimedEffectsFeature:
    """
    Represent `TimedEffectsFeature` object as used by various Hue resources.

    Basic feature containing timed effect properties.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_light_get
    """

    effect: TimedEffectStatus
    status: TimedEffectStatus
    status_values: List[TimedEffectStatus] = field(default_factory=list)
    effect_values: List[TimedEffectStatus] = field(default_factory=list)
    # Duration is mandatory when timed effect is set except for no_effect.
    # Resolution decreases for a larger duration. e.g Effects with duration smaller
    # than a minute will be rounded to a resolution of 1s, while effects with duration
    # larger than an hour will be arounded up to a resolution of 300s.
    # Duration has a max of 21600000 ms.
    duration: Optional[int] = None


@dataclass
class TimedEffectsFeaturePut:
    """
    Represent `TimedEffectsFeature` object when sent to the API in PUT requests.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_light__id__put
    """

    effect: Optional[TimedEffectStatus]
    # Duration is mandatory when timed effect is set except for no_effect.
    # Resolution decreases for a larger duration. e.g Effects with duration smaller
    # than a minute will be rounded to a resolution of 1s, while effects with duration
    # larger than an hour will be arounded up to a resolution of 300s.
    # Duration has a max of 21600000 ms.
    duration: Optional[int] = None


class RecallAction(Enum):
    """Enum with available recall actions."""

    ACTIVE = "active"
    STATIC = "static"
    DYNAMIC_PALETTE = "dynamic_palette"


@dataclass
class RecallFeature:
    """
    Properties to send when calling/setting the `Recall` feature (of a scene) on the api.

    clip-api.schema.json#/definitions/RecallFeature
    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_scene__id__put
    """

    # action: (RecallAction)
    # When writing active, the actions in the scene are executed on the target.
    action: Optional[RecallAction] = None
    # status: (active, dynamic_palette)
    # When writing active, the actions are executed on the target (legacy, use action instead).
    status: Optional[str] = None
    # duration: (integer – maximum: 6000000)
    # transition to the scene within the timeframe given by duration. Accuracy is in 100ms steps.
    duration: Optional[int] = None
    # dimming: (DimmingFeature)
    # override the scene dimming/brightness
    dimming: Optional[DimmingFeatureBase] = None


@dataclass
class PaletteFeatureColor:
    """Represents Color object used in PaletteFeature."""

    color: ColorFeatureBase
    dimming: DimmingFeatureBase


@dataclass
class PaletteFeatureColorTemperature:
    """Represents ColorTemperature object used in PaletteFeature."""

    color_temperature: ColorTemperatureFeatureBase
    dimming: DimmingFeatureBase


@dataclass
class PaletteFeature:
    """
    Group of colors that describe the palette of colors to be used when playing dynamics.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_scene__id__get
    """

    # color: required(array of PaletteFeatureColor – minItems: 0 – maxItems: 9)
    color: List[PaletteFeatureColor]
    # dimming: required(array of DimmingFeature – minItems: 0 – maxItems: 1)
    dimming: List[DimmingFeatureBase]
    # color_temperature: (array of PaletteFeatureColorTemperature – minItems: 0 – maxItems: 1)
    color_temperature: List[PaletteFeatureColorTemperature]


@dataclass
class GradientPoint:
    """Represent a single Gradient (color) Point."""

    color: ColorFeatureBase


class GradientMode:
    """Mode of the Gradient feature."""

    INTERPOLATED_PALETTE = "interpolated_palette"
    INTERPOLATED_PALETTE_MIRRORED = "interpolated_palette_mirrored"
    RANDOM_PIXELATED = "random_pixelated"


@dataclass
class GradientFeatureBase:
    """Represent GradientFeature base properties."""

    # points: Collection of gradients points.
    # For control of the gradient points through a PUT a minimum of 2 points need to be provided.
    points: List[GradientPoint]


@dataclass
class GradientFeature(GradientFeatureBase):
    """
    Represent GradientFeature as used by the lights entities.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_light_get
    """

    # points_capable: required(integer)
    # Number of color points that gradient lamp is capable of showing with gradience.
    points_capable: int
    # mode: Mode in which the points are currently being deployed.
    # If not provided during PUT/POST it will be defaulted to interpolated_palette
    mode: GradientMode = GradientMode.INTERPOLATED_PALETTE
    mode_values: List[GradientMode] = field(default_factory=list)
    pixel_count: Optional[int] = None  # Number of pixels in the device


class Signal(Enum):
    """Enum with various signals."""

    NO_SIGNAL = "no_signal"
    ON_OFF = "on_off"
    UNKNOWN = "unknown"

    @classmethod
    def _missing_(cls: Type, value: object):
        """Set default enum member if an unknown value is provided."""
        return AlertEffectType.UNKNOWN


@dataclass
class SignalingFeatureStatus:
    """Indicates status of active signal. Not available when inactive."""

    signal: Signal
    estimated_end: datetime


@dataclass
class SignalingFeature:
    """Feature containing signaling properties."""

    status: SignalingFeatureStatus


class PowerUpPreset(Enum):
    """Enum with available powerup presets."""

    SAFETY = "safety"
    POWERFAIL = "powerfail"
    LAST_ON_STATE = "last_on_state"
    CUSTOM = "custom"
    UNKNOWN = "unknown"

    @classmethod
    def _missing_(cls: Type, value: object):
        """Set default enum member if an unknown value is provided."""
        return AlertEffectType.UNKNOWN


class PowerUpFeatureOnMode(Enum):
    """Enum with available powerup on modes."""

    ON = "on"
    TOGGLE = "toggle"
    PREVIOUS = "previous"


@dataclass
class PowerUpFeatureOnState:
    """
    State to activate after powerup.

    On will use the value specified in the “on” property.
    When setting mode “on”, the on property must be included.
    Toggle will alternate between on and off on each subsequent power toggle.
    Previous will return to the state it was in before powering off.
    """

    mode: PowerUpFeatureOnMode
    on: Optional[OnFeature] = None


class PowerUpFeatureDimmingMode(Enum):
    """Enum with available powerup dimming modes."""

    DIMMING = "dimming"
    PREVIOUS = "previous"


@dataclass
class PowerUpFeatureDimmingState:
    """
    Dimming will set the brightness to the specified value after power up.

    When setting mode “dimming”, the dimming property must be included.
    Previous will set brightness to the state it was in before powering off.
    """

    mode: PowerUpFeatureDimmingMode
    dimming: Optional[DimmingFeatureBase] = None


class PowerUpFeatureColorMode(Enum):
    """Enum with available powerup color modes."""

    COLOR_TEMPERATURE = "color_temperature"
    COLOR = "color"
    PREVIOUS = "previous"


@dataclass
class PowerUpFeatureColorState:
    """
    Color state to activate after powerup.

    Availability of “color_temperature” and “color” modes depend on the capabilities of the lamp.
    Colortemperature will set the colortemperature to the specified value after power up.
    When setting color_temperature, the color_temperature property must be included Color will
    set the color tot he specified value after power up. When setting color mode,
    the color property must be included Previous will set color to the state
    it was in before powering off.
    """

    mode: PowerUpFeatureColorMode
    color_temperature: Optional[ColorTemperatureFeatureBase] = None
    color: Optional[ColorFeatureBase] = None


@dataclass
class PowerUpFeature:
    """
    Feature containing properties to configure powerup behaviour of a lightsource.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_light_get
    """

    # NOTE: When setting the custom preset the additional properties can be set.
    # For all other presets, no other properties can be included.
    preset: PowerUpPreset
    configured: bool
    on: PowerUpFeatureOnState
    dimming: Optional[PowerUpFeatureDimmingState] = None
    color: Optional[PowerUpFeatureColorState] = None


@dataclass
class PowerUpFeaturePut:
    """
    PowerUp feature properties that can be set/updated with a PUT request.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_light__id__put
    """

    preset: PowerUpPreset
    on: Optional[PowerUpFeatureOnState] = None
    dimming: Optional[PowerUpFeatureDimmingState] = None
    color: Optional[PowerUpFeatureColorState] = None



"""
Model(s) for light_level resource on HUE bridge.

https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_light_level
"""
from dataclasses import dataclass
from typing import Optional

from .resource import ResourceIdentifier, ResourceTypes


@dataclass
class LightLevelFeature:
    """Represent LightLevel Feature used by Lightlevel resources."""

    light_level: int
    light_level_valid: bool


@dataclass
class LightLevel:
    """
    Represent a (full) `LightLevel` resource when retrieved from the api.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_light_level_get
    """

    id: str
    owner: ResourceIdentifier
    enabled: bool
    light: LightLevelFeature

    id_v1: Optional[str] = None
    type: ResourceTypes = ResourceTypes.LIGHT_LEVEL


@dataclass
class LightLevelPut:
    """
    LightLevel resource properties that can be set/updated with a PUT request.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_light_level__id__put
    """

    enabled: Optional[bool] = None



"""
Model(s) for bridge_home resource on HUE bridge.

https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_bridge_home
"""
from dataclasses import dataclass
from typing import List, Optional

from .resource import ResourceIdentifier, ResourceTypes


@dataclass
class BridgeHome:
    """
    Represent the (full) `BridgeHome` object as retrieved from the Hue api.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_bridge_home_get
    """

    id: str
    # services: required(array of ResourceIdentifier)
    # References all services aggregating control and state of children in the group
    # This includes all services grouped in the group hierarchy given by child relation
    # This includes all services of a device grouped in the group hierarchy given by child relation
    # Aggregation is per service type, ie every service type which can be grouped has a
    # corresponding definition of grouped type
    # Supported types “light”
    services: List[ResourceIdentifier]
    children: List[ResourceIdentifier]

    id_v1: Optional[str] = None
    type: ResourceTypes = ResourceTypes.BRIDGE_HOME



"""
Model(s) for Smart Scene resource on HUE bridge.

https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_smart_scene
"""
from dataclasses import dataclass
from enum import Enum
from typing import List, Optional, Type

from .resource import ResourceIdentifier, ResourceTypes
from .scene import SceneMetadata, SceneMetadataPut


class WeekDay(Enum):
    """Represent Day of Week in TimeSlot."""

    SUNDAY = "sunday"
    MONDAY = "monday"
    TUESDAY = "tuesday"
    WEDNESDAY = "wednesday"
    THURSDAY = "thursday"
    FRIDAY = "friday"
    SATURDAY = "saturday"


class SmartSceneState(Enum):
    """The state of the Active Scene."""

    ACTIVE = "active"
    INACTIVE = "inactive"

    @classmethod
    def _missing_(cls: Type, value: object):
        """Set default enum member if an unknown value is provided."""
        return SmartSceneState.INACTIVE


@dataclass
class TimeslotStartTimeTime:
    """Time object."""

    hour: int  # minimum: 0 – maximum: 23
    minute: int  # minimum: 0 – maximum: 59
    second: int  # minimum: 0 – maximum: 59

    def __post_init__(self):
        """Validate values."""
        if self.hour < 0 or self.hour > 23:
            raise ValueError("Hour must be a value within range of 0 and 23")
        if self.minute < 0 or self.minute > 59:
            raise ValueError("Minute must be a value within range of 0 and 59")
        if self.second < 0 or self.second > 59:
            raise ValueError("Second must be a value within range of 0 and 59")


@dataclass
class TimeslotStartTime:
    """Representation of a Start time object within a timeslot."""

    kind: str  # currently fixed to 'time'
    time: TimeslotStartTimeTime


@dataclass
class SmartSceneTimeslot:
    """
    Represent SmartSceneTimeslot as used by Smart Scenes.

    Information on what is the light state for every timeslot of the day.
    """

    start_time: TimeslotStartTime
    target: ResourceIdentifier


@dataclass
class DayTimeSlots:
    """Represent DayTimeSlots information, used by Smart Scenes."""

    timeslots: List[SmartSceneTimeslot]
    recurrence: List[WeekDay]


@dataclass
class SmartSceneActiveTimeslot:
    """The active time slot in execution."""

    timeslot_id: int
    weekday: WeekDay


class SmartSceneRecallAction:
    """
    Enum with possible recall actions for smart scenes.

    Activate will start the smart (24h) scene; deactivate will stop it.
    """

    ACTIVATE = "activate"
    DEACTIVATE = "deactivate"


@dataclass
class SmartSceneRecall:
    """Properties to send when activating a Smart Scene."""

    action: SmartSceneRecallAction


@dataclass
class SmartScene:
    """
    Represent (full) `SmartScene` Model when retrieved from the API.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_smart_scene__id__get
    """

    id: str
    metadata: SceneMetadata
    # group: required(object)
    # Group associated with this Scene. All services in the group are part of this scene.
    # If the group is changed the scene is updated (e.g. light added/removed)
    group: ResourceIdentifier
    # actions: required(array of Action)
    # List of actions to be executed synchronously on recal
    week_timeslots: List[DayTimeSlots]
    # state: the current state of the smart scene.
    # The default state is inactive if no recall is provided
    state: SmartSceneState

    # optional params

    # active_timeslot: information on what is the light state for every timeslot of the day
    active_timeslot: Optional[SmartSceneActiveTimeslot] = None
    id_v1: Optional[str] = None

    type: ResourceTypes = ResourceTypes.SMART_SCENE


@dataclass
class SmartScenePut:
    """
    Properties to send when updating/setting a `SmartScene` object on the api.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_smart_scene__id__put
    """

    metadata: Optional[SceneMetadataPut] = None
    week_timeslots: Optional[List[DayTimeSlots]] = None
    recall: Optional[SmartSceneRecall] = None


@dataclass
class SmartSceneScenePost:
    """
    Properties to send when creating a `SmartScene` object on the api.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_smarft_scene_post
    """

    metadata: SceneMetadata
    group: ResourceIdentifier
    week_timeslots: List[DayTimeSlots]
    recall: Optional[SmartSceneRecall] = None
    type: ResourceTypes = ResourceTypes.SMART_SCENE



"""Controller/model for Hue resource of type Group."""
from __future__ import annotations
from logging import Logger
from typing import Any, Coroutine, Dict, List, TypedDict

from .api import APIItems


class Groups(APIItems):
    """
    Represents Hue Groups (zone/room).

    https://developers.meethue.com/documentation/groups-api
    """

    def __init__(self, logger: Logger, raw: Dict[str, Any], request: Coroutine) -> None:
        """Initialize instance."""
        super().__init__(logger, raw, request, "groups", Group)

    async def get_all_lights_group(self) -> "Group":
        """Return special all lights group."""
        return Group("0", await self._request("get", "groups/0"), self._request)


class GroupState(TypedDict):
    """Represents state dict for a group."""

    all_on: bool
    any_on: bool


class GroupAction(TypedDict, total=False):
    """Represents action dict for a group."""

    on: bool
    bri: int  # optional
    hue: int  # optional
    sat: int  # optional
    effect: str  # optional
    xy: List[float]  # optional
    ct: int  # optional
    alert: str
    colormode: str  # optional


class Group:
    """Represents a Hue Group."""

    ITEM_TYPE = "groups"

    def __init__(self, id: str, raw: Dict[str, Any], request: Coroutine) -> None:
        """Initialize instance."""
        self.id = id
        self.raw = raw
        self._request = request

    @property
    def type(self) -> str:
        """Return type of the group (Zone or Room)."""
        return self.raw["type"]

    @property
    def name(self) -> str:
        """Return name."""
        return self.raw["name"]

    @property
    def uniqueid(self) -> str:
        """
        Return Unique ID for the group.

        Requires API version 1.9+ and only for groups that are the type
        Luminaire or Lightsource.
        """
        return self.raw.get("uniqueid")

    @property
    def action(self) -> GroupAction:
        """Return current group action."""
        return GroupAction(self.raw["action"])

    @property
    def state(self) -> GroupState:
        """Return current group state."""
        return GroupState(self.raw["state"])

    @property
    def lights(self) -> List[str]:
        """Return id's of lights that are memebers of this group."""
        return self.raw["lights"]

    @property
    def sensors(self) -> List[str]:
        """Return id's of sensors that are memebers of this group."""
        return self.raw["sensors"]

    async def set_action(
        self,
        on=None,
        bri=None,
        hue=None,
        sat=None,
        xy=None,
        ct=None,
        alert=None,
        effect=None,
        transitiontime=None,
        bri_inc=None,
        sat_inc=None,
        hue_inc=None,
        ct_inc=None,
        xy_inc=None,
        scene=None,
    ) -> None:
        """Change action of a group."""
        data = {
            key: value
            for key, value in {
                "on": on,
                "bri": bri,
                "hue": hue,
                "sat": sat,
                "xy": xy,
                "ct": ct,
                "alert": alert,
                "effect": effect,
                "transitiontime": transitiontime,
                "bri_inc": bri_inc,
                "sat_inc": sat_inc,
                "hue_inc": hue_inc,
                "ct_inc": ct_inc,
                "xy_inc": xy_inc,
                "scene": scene,
            }.items()
            if value is not None
        }

        await self._request("put", "groups/{}/action".format(self.id), json=data)



"""
Model(s) for entertainment_configuration resource on HUE bridge.

https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_entertainment_configuration
"""
from dataclasses import dataclass
from enum import Enum
from typing import List, Optional

from .feature import Position
from .resource import ResourceIdentifier, ResourceTypes


@dataclass
class Segment:
    """
    Represent a Segment object.

    Contains the segments configuration of the device for entertainment purposes.
    A device can be segmented in a single way.
    """

    length: int
    start: int


@dataclass
class SegmentReference:
    """
    Represent a SegmentReference object.

    From which entertainment_service, which segment is a member of this channel.
    """

    service: ResourceIdentifier
    index: int


@dataclass
class EntertainmentChannel:
    """
    Represent a EntertainmentChannel object as used by the Hue api.

    Holds position of the channel and list of members,
    which is an array of segments resource identifiers.
    """

    channel_id: int
    # xyz position of this channel. It is the average position of its members.
    position: Position
    members: List[SegmentReference]


class EntertainmentConfigurationType(Enum):
    """Enum with possible Entertainment Configuration Types."""

    SCREEN = "screen"  # Channels are organized around content from a screen
    MUSIC = "music"  # Channels are organized for music synchronization
    THREEDEESPACE = "3dspace"  # Channels are organized to provide 3d spacial effects
    OTHER = "other"  # General use case "
    MONITOR = "monitor"  # Channels are organized around content from monitors


class EntertainmentStatus(Enum):
    """Enum with possible Entertainment Statuses."""

    ACTIVE = "active"
    INACTIVE = "inactive"


class StreamingProxyMode(Enum):
    """Enum with possible StreamingProxy Modes."""

    AUTO = "auto"
    MANUAL = "manual"


@dataclass
class StreamingProxy:
    """Represent a StreamingProxy object as used by the Hue api."""

    mode: StreamingProxyMode
    node: ResourceIdentifier


@dataclass
class ServiceLocation:
    """Represent a ServiceLocation object as used by the Hue api."""

    service: ResourceIdentifier
    positions: List[Position]
    position: Optional[Position] = None


@dataclass
class EntertainmentLocations:
    """
    Represent a EntertainmentLocations object as used by the Hue api.

    The position of each entertainment service in the configuration.
    If the service has several segments, more than one x,y,z location may be needed
    """

    service_locations: List[ServiceLocation]


class EntertainmentConfigurationAction(Enum):
    """
    Enum with possible Entertainment Configuration Actions.

    as defined in:
    clip-api.schema.json#/definitions/EntertainmentConfigurationPut
    """

    START = "start"
    STOP = "stop"


@dataclass
class EntertainmentConfigurationMetaData:
    """Represent EntertainmentConfigurationMetaData for a device object as used by the Hue api."""

    name: str


@dataclass
class EntertainmentConfiguration:
    """
    Represent a (full) `EntertainmentConfiguration` resource when retrieved from the api.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_entertainment_configuration_get
    """

    id: str
    metadata: EntertainmentConfigurationMetaData
    configuration_type: EntertainmentConfigurationType
    status: EntertainmentStatus
    stream_proxy: StreamingProxy
    channels: List[EntertainmentChannel]
    locations: EntertainmentLocations
    light_services: Optional[List[ResourceIdentifier]] = None

    active_streamer: Optional[ResourceIdentifier] = None
    id_v1: Optional[str] = None
    type: ResourceTypes = ResourceTypes.ENTERTAINMENT_CONFIGURATION



"""
Model(s) for device resource on HUE bridge.

https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_device
"""
from dataclasses import dataclass
from enum import Enum
from typing import List, Optional, Set, Type

from .feature import IdentifyFeature
from .resource import SENSOR_RESOURCE_TYPES, ResourceIdentifier, ResourceTypes


class DeviceArchetypes(Enum):
    """Enum with all possible Device archtypes."""

    BRIDGE_V2 = "bridge_v2"
    UNKNOWN_ARCHETYPE = "unknown_archetype"
    CLASSIC_BULB = "classic_bulb"
    SULTAN_BULB = "sultan_bulb"
    FLOOD_BULB = "flood_bulb"
    SPOT_BULB = "spot_bulb"
    CANDLE_BULB = "candle_bulb"
    LUSTER_BULB = "luster_bulb"
    PENDANT_ROUND = "pendant_round"
    PENDANT_LONG = "pendant_long"
    CEILING_ROUND = "ceiling_round"
    CEILING_SQUARE = "ceiling_square"
    FLOOR_SHADE = "floor_shade"
    FLOOR_LANTERN = "floor_lantern"
    TABLE_SHADE = "table_shade"
    RECESSED_CEILING = "recessed_ceiling"
    RECESSED_FLOOR = "recessed_floor"
    SINGLE_SPOT = "single_spot"
    DOUBLE_SPOT = "double_spot"
    TABLE_WASH = "table_wash"
    WALL_LANTERN = "wall_lantern"
    WALL_SHADE = "wall_shade"
    FLEXIBLE_LAMP = "flexible_lamp"
    GROUND_SPOT = "ground_spot"
    WALL_SPOT = "wall_spot"
    PLUG = "plug"
    HUE_GO = "hue_go"
    HUE_LIGHTSTRIP = "hue_lightstrip"
    HUE_IRIS = "hue_iris"
    HUE_BLOOM = "hue_bloom"
    BOLLARD = "bollard"
    WALL_WASHER = "wall_washer"
    HUE_PLAY = "hue_play"
    VINTAGE_BULB = "vintage_bulb"
    CHRISTMAS_TREE = "christmas_tree"
    STRING_LIGHT = "string_light"
    HUE_CENTRIS = "hue_centris"
    HUE_LIGHTSTRIP_TV = "hue_lightstrip_tv"
    HUE_LIGHTSTRIP_PC = "hue_lightstrip_pc"
    HUE_TUBE = "hue_tube"
    HUE_SIGNE = "hue_signe"
    PENDANT_SPOT = "pendant_spot"
    CEILING_HORIZONTAL = "ceiling_horizontal"
    CEILING_TUBE = "ceiling_tube"

    @classmethod
    def _missing_(cls: Type, value: object):
        """Set default enum member if an unknown value is provided."""
        return DeviceArchetypes.UNKNOWN_ARCHETYPE


@dataclass
class DeviceProductData:
    """Represent a DeviceProductData object as used by the Hue api."""

    model_id: str
    manufacturer_name: str
    product_name: str
    product_archetype: DeviceArchetypes
    certified: bool
    software_version: str
    # Hardware type; identified by Manufacturer code and ImageType
    hardware_platform_type: Optional[str] = None


@dataclass
class DeviceMetaData:
    """Represent MetaData for a device object as used by the Hue api."""

    archetype: DeviceArchetypes
    name: str


@dataclass
class DeviceMetaDataPut:
    """Represent MetaData for a device object on update/PUT."""

    archetype: Optional[DeviceArchetypes]
    name: Optional[str]


@dataclass
class Device:
    """
    Represent a (full) `Device` resource as retrieved from the Hue api.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_device_get
    """

    id: str
    # services: required(array of ResourceIdentifierGet)
    # References all services aggregating control and state of children in the group
    # This includes all services grouped in the group hierarchy given by child relation
    # This includes all services of a device grouped in the group hierarchy given by child relation
    # Aggregation is per service type, ie every service type which can be grouped has a
    # corresponding definition of grouped type
    # Supported types “light”
    services: List[ResourceIdentifier]
    product_data: DeviceProductData
    metadata: DeviceMetaData

    id_v1: Optional[str] = None
    type: ResourceTypes = ResourceTypes.DEVICE

    @property
    def lights(self) -> Set[str]:
        """Return a set of light id's belonging to this group/device."""
        return {x.rid for x in self.services if x.rtype == ResourceTypes.LIGHT}

    @property
    def sensors(self) -> Set[str]:
        """Return a set of sensor id's belonging to this group/device."""
        return {x.rid for x in self.services if x.rtype in SENSOR_RESOURCE_TYPES}


@dataclass
class DevicePut:
    """
    Device resource properties that can be set/updated with a PUT request.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_device__id__put
    """

    metadata: Optional[DeviceMetaDataPut] = None
    identify: Optional[IdentifyFeature] = None



"""Generic/base Resource Model(s)."""

from dataclasses import dataclass
from enum import Enum
from typing import Type


class ResourceTypes(Enum):
    """
    Type of the supported resources.

    clip-api.schema.json#/definitions/ResourceTypes
    """

    DEVICE = "device"
    BRIDGE_HOME = "bridge_home"
    ROOM = "room"
    ZONE = "zone"
    LIGHT = "light"
    BUTTON = "button"
    RELATIVE_ROTARY = "relative_rotary"
    TEMPERATURE = "temperature"
    LIGHT_LEVEL = "light_level"
    MOTION = "motion"
    ENTERTAINMENT = "entertainment"
    GROUPED_LIGHT = "grouped_light"
    DEVICE_POWER = "device_power"
    DEVICE_UPDATE = "device_update"
    IP_CONNECTIVITY = "ip_connectivity"
    ZIGBEE_BRIDGE_CONNECTIVITY = "zigbee_bridge_connectivity"
    ZIGBEE_CONNECTIVITY = "zigbee_connectivity"
    ZGP_CONNECTIVITY = "zgp_connectivity"
    REMOTE_ACCESS = "remote_access"
    BRIDGE = "bridge"
    DEVICE_DISCOVERY = "device_discovery"
    SYSTEM_UPDATE = "system_update"
    SCENE = "scene"
    SMART_SCENE = "smart_scene"
    ENTERTAINMENT_CONFIGURATION = "entertainment_configuration"
    PUBLIC_IMAGE = "public_image"
    AUTH_V1 = "auth_v1"
    BEHAVIOR_SCRIPT = "behavior_script"
    BEHAVIOR_INSTANCE = "behavior_instance"
    GEOFENCE = "geofence"
    GEOFENCE_CLIENT = "geofence_client"
    DEPENDER = "depender"
    HOMEKIT = "homekit"
    MATTER = "matter"
    MATTER_FABRIC = "matter_fabric"
    UNKNOWN = "unknown"

    @classmethod
    def _missing_(cls: Type, value: object):
        """Set default enum member if an unknown value is provided."""
        return ResourceTypes.UNKNOWN


SENSOR_RESOURCE_TYPES = (
    ResourceTypes.DEVICE_POWER,
    ResourceTypes.BUTTON,
    ResourceTypes.GEOFENCE_CLIENT,
    ResourceTypes.LIGHT_LEVEL,
    ResourceTypes.MOTION,
    ResourceTypes.RELATIVE_ROTARY,
    ResourceTypes.TEMPERATURE,
    ResourceTypes.ZIGBEE_CONNECTIVITY,
)


@dataclass
class ResourceIdentifier:
    """
    Represent a ResourceIdentifier object as used by the Hue api.

    clip-api.schema.json#/definitions/ResourceIdentifierGet
    clip-api.schema.json#/definitions/ResourceIdentifierPost
    clip-api.schema.json#/definitions/ResourceIdentifierPut
    clip-api.schema.json#/definitions/ResourceIdentifierDelete
    """

    rid: str  # UUID
    rtype: ResourceTypes



"""
Model(s) for behavior_script resource on HUE bridge.

API to discover available scripts that can be instantiated
https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_behavior_script
"""
from dataclasses import dataclass
from enum import Enum
from typing import List, Optional, Type

from .resource import ResourceTypes


class BehaviorScriptCategory(Enum):
    """Enum with various Bahavior Script Categories."""

    AUTOMATION = "automation"
    ENTERTAINMENT = "entertainment"
    ACCESSORY = "accessory"
    OTHER = "other"

    @classmethod
    def _missing_(cls: Type, value: object):
        """Set default enum member if an unknown value is provided."""
        return BehaviorScriptCategory.OTHER


@dataclass
class BehaviorScriptMetadata:
    """Represent BehaviorScript Metadata object as used by BehaviorScript resource."""

    name: Optional[str] = None
    category: BehaviorScriptCategory = BehaviorScriptCategory.OTHER


@dataclass
class BehaviorScript:
    """
    Represent a (full) `BehaviorScript` resource when retrieved from the api.

    Available scripts that can be instantiated.
    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_behavior_script_get
    """

    id: str
    description: str
    # configuration_schema: required(object)
    # JSON schema object used for validating ScriptInstance.configuration property.
    configuration_schema: dict
    # trigger_schema: required(object)
    # JSON schema object used for validating ScriptInstance.trigger property.
    trigger_schema: dict
    # state_schema: required(object)
    # JSON schema of ScriptInstance.state property.
    state_schema: dict
    version: str
    metadata: BehaviorScriptMetadata
    supported_features: Optional[List[str]]
    max_number_instances: Optional[int] = None

    id_v1: Optional[str] = None
    type: ResourceTypes = ResourceTypes.BEHAVIOR_SCRIPT



"""
Model(s) for homekit resource on HUE bridge.

https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_homekit
"""
from dataclasses import dataclass
from enum import Enum
from typing import Optional

from .resource import ResourceTypes


class HomekitStatus(Enum):
    """Enum with the possible homekit status."""

    PAIRING = "pairing"
    PAIRED = "paired"
    UNPAIRED = "unpaired"


class HomekitAction(Enum):
    """Enum with the possible homekit status."""

    NONE = ""
    RESET = "reset"
    # reset: Reset homekit, including removing all pairings and reset state and Bonjour service
    # to factory settings. The Homekit will start functioning after approximately 10 seconds.


@dataclass
class Homekit:
    """
    Represent a (full) `Homekit` resource when retrieved from the api.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_homekit_get
    """

    id: str
    status: HomekitStatus

    id_v1: Optional[str] = None
    type: ResourceTypes = ResourceTypes.HOMEKIT


@dataclass
class HomekitPut:
    """
    Homekit resource properties that can be set/updated with a PUT request.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_homekit__id__put
    """

    action: Optional[HomekitAction] = None



"""
Model(s) for device_power resource on HUE bridge.

https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_device_power
"""
from dataclasses import dataclass
from enum import Enum


from typing import Optional

from .resource import ResourceIdentifier, ResourceTypes


class BatteryState(Enum):
    """
    Enum with all possible BatteryStates.

    - normal: battery level is sufficient
    - low: battery level low, some features might stop working, please change battery soon
    - critical: battery level critical, device can fail any moment
    """

    NORMAL = "normal"
    LOW = "low"
    CRITICAL = "critical"


@dataclass
class PowerState:
    """Represent PowerState as retrieved from api."""

    battery_level: Optional[int]
    battery_state: Optional[BatteryState]


@dataclass
class DevicePower:
    """
    Represent a (full) `DevicePower` resource when retrieved from the api.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_device_power_get
    """

    id: str
    owner: ResourceIdentifier
    power_state: PowerState

    id_v1: Optional[str] = None
    type: ResourceTypes = ResourceTypes.DEVICE_POWER



"""
Model(s) for relative_rotary resource on HUE bridge.

https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_relative_rotary
"""
from dataclasses import dataclass
from enum import Enum
from typing import Optional, Type

from .resource import ResourceIdentifier, ResourceTypes


class RelativeRotaryAction(Enum):
    """Enum with possible relative_rotary actions."""

    START = "start"
    REPEAT = "repeat"
    UNKNOWN = "unknown"

    @classmethod
    def _missing_(cls: Type, value: object):
        """Set default enum member if an unknown value is provided."""
        return RelativeRotaryAction.UNKNOWN


class RelativeRotaryDirection(Enum):
    """Enum with possible rotary directions."""

    CLOCK_WISE = "clock_wise"
    COUNTER_CLOCK_WISE = "counter_clock_wise"


@dataclass
class RelativeRotaryRotation:
    """
    Represent Rotation object as used by the Hue api.

    direction: required(one of clock_wise, counter_clock_wise)
    A rotation opposite to the previous rotation, will always start with new start command.

    steps: required(integer - minimum: 0 - maximum: 32767)
    amount of rotation since previous event in case of repeat,
    amount of rotation since start in case of a start_event.
    Resolution = 1000 steps / 360 degree rotation.

    duration: required(integer - minimum: 0 - maximum: 65534)
    duration of rotation since previous event in case of repeat,
    amount of rotation since start in case of a start_event. duration is specified in miliseconds.
    """

    direction: RelativeRotaryDirection
    duration: int
    steps: int


@dataclass
class RelativeRotaryEvent:
    """Represent RelativeRotaryEvent object as used by the Hue api."""

    action: RelativeRotaryAction
    rotation: RelativeRotaryRotation


@dataclass
class RelativeRotaryFeature:
    """Represent RelativeRotaryFeature object as used by the Hue api."""

    last_event: RelativeRotaryEvent


@dataclass
class RelativeRotary:
    """
    Represent a (full) `RelativeRotary` resource when retrieved from the api.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_relative_rotary_get
    """

    id: str
    owner: ResourceIdentifier

    relative_rotary: Optional[RelativeRotaryFeature] = None
    id_v1: Optional[str] = None
    type: ResourceTypes = ResourceTypes.RELATIVE_ROTARY
